"""entry point for the paragon CLI."""
# pylint: disable = no-value-for-parameter
from typing import List
import click
from paragon import Paragon
from paragon.mark import Mark
from paragon.utils import Utils


@click.command()
@click.argument("code", nargs=-1)
@click.option(
    "--accuracy",
    "-a",
    required=False,
    default=1,
    help="Number of iterations",
)
@click.option(
    "--files",
    "-f",
    required=False,
    multiple=True,
    help="Input file path(s)",
    type=click.Path(),
)
@click.option(
    "--output",
    "-o",
    required=False,
    nargs=1,
    help="Output file path",
    type=click.Path(),
)
def cli(code: List[str], accuracy: int, files: List[str], output: str):
    """entry point for the paragon CLI."""

    # nothing to benchmark
    if len(code) == 0 and len(files) == 0:
        click.echo("You must provide python code to benchmark.", err=True)
        return

    start, current_benchmark = Mark(), 1
    for val in code:
        click.secho(f"Benchmark #{current_benchmark}", fg="white", bold=True)
        try:
            Paragon.bench(code=val, accuracy=accuracy)
        except (NameError, SyntaxError) as error:
            click.echo(f"Error: {error}", err=True)
        current_benchmark += 1

    if len(files):
        for file in files:
            res, status = Utils.verify_file(file)
            click.secho(f"Benchmark #{current_benchmark}", fg="white", bold=True)

            if not status:
                click.echo(res + "\n", err=True)
                continue

            try:
                Paragon.bench(code=res, accuracy=accuracy)
            except (NameError, SyntaxError) as error:
                click.echo(f"Error: {error}", err=True)

            current_benchmark += 1

    if output:
        print(output)

    print(f"✨ Done in {start.diff()} s ✨")


if __name__ == "__main__":
    cli()



"""all benchmark related logic"""
# pylint: disable = exec-used

import click
from paragon.stats import Stats
from paragon.utils import Utils
from paragon.mark import Mark
from paragon.animate import Animate


class Paragon:
    """handles code benchmarking"""

    @staticmethod
    def benchmark(name: str = None, accuracy: int = 1):
        """benchmark decorator
        - retrieves decoratored functions name and args
        - calls bench with the function call and scope

        :param name: name of benchmark
        :accuracy: number of iterations
        """

        def inner(func):
            def wrap(*args):
                nonlocal name

                start, func_name = Mark(), func.__name__

                click.secho(
                    "Benchmark: {}".format(name or func_name),
                    fg="white",
                    bold=True,
                )

                # gather locals and build code string
                env = {func_name: func, args: args}
                code = f"{func_name}(*{args})"

                # run bench
                Paragon.bench(code=code, env=env, accuracy=accuracy)
                print(f"✨ Done in {start.diff()} s ✨\n")

                return func(*args)

            return wrap

        return inner

    @staticmethod
    def bench(code: str, env: object = None, accuracy: int = 1):
        """benchmark function
        - benches passed in code, taking in optional locals and number
        of iterations

        :param code: code to benchmark
        :accuracy: number of iterations
        :env: locals
        """
        # don't need stdout
        Utils.redirect_stdout()

        # run once to make sure it's valid python
        err, status = Utils.run_once(code, env)
        if not status:
            raise err

        mark, animate, times = Mark(), Animate(accuracy), []

        # begin cycle
        for _ in range(accuracy):
            exec(code, globals(), env or globals())
            animate.next()
            times.append(mark.diff())
            mark.reset()

        # cleanup animation and output stats
        animate.done()
        Stats(times).output()



"""all statistics related logic"""
import click
from paragon.utils import Utils


class Stats:
    """handles statistics compute and output"""

    def __init__(self, times):
        self.times = times

    @property
    def avg(self):
        """returns the average time"""
        return sum(self.times) / len(self.times)

    @property
    def range(self):
        """returns the range (min, max) of all times"""
        return (min(self.times), max(self.times))

    def standard_deviation(self):
        """calculates the standard deviation from the mean"""
        return round(
            (sum([((time - self.avg) ** 2) for time in self.times]) / len(self.times))
            ** 0.5,
            10,
        )

    def output(self):
        """prints stats to stdout"""
        Utils.reset_stdout()

        print(
            "\nTime ({} \xB1 {}): {} s \xB1 {} s.".format(
                click.style("mean", fg="green"),
                click.style("\u03C3", fg="green"),
                click.style("{:.2f}".format(self.avg), fg="green"),
                click.style("{:.2f}".format(self.standard_deviation()), fg="green"),
            )
        )

        print(
            "Range ({} ... {}): {} s ... {} s.\n".format(
                click.style("min", fg="blue"),
                click.style("max", fg="red"),
                click.style("{:.2f}".format(self.range[0]), fg="blue"),
                click.style("{:.2f}".format(self.range[1]), fg="red"),
            )
        )



"""all utility related logic"""
# pylint: disable = exec-used

import os
import sys


class Utils:
    """utility methods"""

    @staticmethod
    def reset_stdout():
        """set stdout back to normal"""
        sys.stdout = sys.__stdout__

    @staticmethod
    def redirect_stdout(redirect_value=None):
        """set stdout to provided value or None"""
        sys.stdout = redirect_value

    @staticmethod
    def verify_file(file):
        """verifies a python file to benchmark"""
        if not os.path.exists(file):
            return ("File does not exist.", False)

        _, ext = os.path.splitext(file)
        if ext != ".py":
            return ("File must be a python file.", False)

        return (open(file, "r").read(), True)

    @staticmethod
    def run_once(code, env: object = None):
        """run code once to make sure it's valid"""
        try:
            exec(code, globals(), env or globals())
            return (None, True)
        except (NameError, SyntaxError) as error:
            Utils.reset_stdout()
            return (error, False)
